/*
 * Project: SIP Registrar - An example SIP Registrar implementing the chapter 10 of 
 * 			RFC 3261 describing the behaviour of a SIP Registrar. The SIP Registrar
 * 			is implemented using SIP Servlet technologies. 
 *
 * File: Registrar.java
 * Author: Michael Maretzke
 * License: Distributable under LGPL license - see terms of license at gnu.org
 * Created: 22nd November 2006, 22:17
 * Version: 1.0
 */
package com.maretzke.sip.tools.proxy;

import com.maretzke.sip.tools.DebugSipServlet;
import com.maretzke.sip.tools.registrar.QueryLocationService;
import com.maretzke.sip.tools.registrar.Registration;
import com.maretzke.sip.tools.registrar.Contact;

import java.io.IOException;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.sip.TooManyHopsException;
import javax.servlet.sip.SipFactory;
import javax.servlet.sip.SipServletRequest;
import javax.servlet.sip.Proxy;
import javax.servlet.sip.URI;

/**
 * The SIP Servlet LookupProxy implements typical proxy server capabilities.
 * Here, the implementation combines with the SIP Registrar project and does a
 * lookup in the binding information stored at InMemoryLocationService. The
 * interface used is QueryLocationService.
 * 
 * The implementation is straight forward. If the SIP INVITE request does belong
 * to the domain the LookupProxy Servlet serves then it will lookup and resolve
 * the AOR of the request. Otherwise, LookupProxy will return a 404 Not Found
 * error message.
 * 
 * @author Michael Maretzke
 * 
 */
public class LookupProxy extends DebugSipServlet {

	// unique identifier for this class
	private static final long serialVersionUID = 6548923479L;

	// string to store the implementation of the LocationService implementation.
	// the information is stored actually in the JNDI tree for further
	// reference.
	private final String JNDI_LOCATION_SERVICE = "java:comp/com/maretzke/sip/tools/registrar/locationService";

	// the implementation of the LocationService
	private QueryLocationService queryLocationService;

	// the domain served by this Registrar - stored in the sip.xml descriptor
	private String domain;

	// a reference to the SipFactory used by the Registrar
	private SipFactory sipFactory;

	/**
	 * Initializes the SIP Servlet. Called only through initialization of the
	 * SIP Servlet.
	 * 
	 * @param servletConfig information provided by the application server
	 */
	public void init(ServletConfig servletConfig) throws ServletException {
		// call the parents' init method
		super.init(servletConfig);

		// get a reference to the SipFactory
		sipFactory = (SipFactory) getServletContext().getAttribute(SIP_FACTORY);

		// load the domain to be served by this server from the sip.xml
		// descriptor
		domain = getServletConfig().getInitParameter("LookupProxy.Domain");
		if (domain == null)
			domain = "default";

		trace(INFO, "LookupProxy initialized. Serving domain: " + domain + ".");
		queryLocationService = null;
	}

	/**
	 * Destroys the SIP Servlet. Called by the application server environment.
	 */
	public void destroy() {
		trace(INFO, "LookupProxy is shutting down now.");
	}

	/**
	 * Invoked by the application server upon the receipt of a SIP INVITE
	 * message.
	 * 
	 * @param req the SipServletRequest generated by the application server
	 */
	public void doInvite(SipServletRequest req) throws ServletException, IOException {

		// lookup the LocationService from JNDI
		queryLocationService = getQueryLocationService();

		trace(INFO, "------------------------------------------");
		trace(INFO, "LookupProxy received a SIP INVITE message.");

		URI aor = req.getTo().getURI();

		Registration reg = queryLocationService.lookupBinding(aor.toString());
		if (reg == null) {
			// not found
			trace(INFO, "AOR is not registered at this domain.");
			sendErrorResponse(req, 404, "Not Found - AOR is not registered.");
			return;

		} else {
			Contact contact = (Contact) reg.getContacts().iterator().next();
			URI contactURI = sipFactory.createURI(contact.getContact());

			trace(INFO, "Proxying the request for " + aor + " to " + contactURI);
			trace(INFO, "------------------------------------------");

			// proxy the request
			try {
				Proxy proxy = req.getProxy(true);
				proxy.setRecordRoute(false);
				proxy.setRecurse(true);
				proxy.setSupervised(false);
				proxy.proxyTo(contactURI);
			} catch (TooManyHopsException e) {
				e.printStackTrace();
				throw new ServletException(e);
			}
		}
	}

	/**
	 * Lookup the implementation of the interface QueryLocationService in the
	 * JNDI environment of the application server.
	 * 
	 * @return the implementation of the interface LocationService
	 */
	private QueryLocationService getQueryLocationService() {
		Context ctx;
		try {
			ctx = new InitialContext();
			return (QueryLocationService) ctx.lookup(JNDI_LOCATION_SERVICE);
		} catch (NamingException e) {
			e.printStackTrace();
			return null;
		}
	}
}
